from __future__ import annotations

import builtins
import sys

import pytest

from orng import ArrayRNG
from orng._utils import normalize_shape, total_size
from orng.backends import _FACTORIES
from orng.backends import numpy as numpy_backend


def test_normalize_shape_accepts_valid_inputs():
    assert normalize_shape(None) == ()
    assert normalize_shape(5) == (5,)
    assert normalize_shape([2, 3]) == (2, 3)


def test_normalize_shape_rejects_negative_dimensions():
    with pytest.raises(ValueError):
        normalize_shape(-1)
    with pytest.raises(ValueError):
        normalize_shape([2, -1])


def test_total_size_handles_scalar_and_shapes():
    assert total_size(()) == 1
    assert total_size((3,)) == 3
    assert total_size((2, 4)) == 8


def test_array_rng_delegates_to_backends(monkeypatch):
    class StubBackend:
        def __init__(self, *, seed, generator, device):
            self.seed = seed
            self.generator = generator
            self.device = device
            self.calls: list[tuple[str, tuple]] = []

        def random(self, *, size, dtype):
            self.calls.append(("random", size, dtype))
            return ("random", size, dtype, self.seed)

        def uniform(self, *, low, high, size, dtype):
            self.calls.append(("uniform", low, high, size, dtype))
            return ("uniform", low, high, size, dtype, self.seed)

        def normal(self, *, loc, scale, size, dtype):
            self.calls.append(("normal", loc, scale, size, dtype))
            return ("normal", loc, scale, size, dtype, self.seed)

        def gamma(self, *, shape, scale, size, dtype):
            self.calls.append(("gamma", shape, scale, size, dtype))
            return ("gamma", shape, scale, size, dtype, self.seed)

        def choice(self, population, *, size, replace, probabilities):
            self.calls.append(
                ("choice", population, size, replace, probabilities)
            )
            return (
                "choice",
                population,
                size,
                replace,
                probabilities,
                self.seed,
            )

    instances: list[StubBackend] = []

    def fake_factory(**kwargs):
        backend = StubBackend(**kwargs)
        instances.append(backend)
        return backend

    monkeypatch.setitem(_FACTORIES, "numpy", fake_factory)

    rng = ArrayRNG(backend="numpy", seed=7, generator="sentinel", device="cpu")
    assert len(instances) == 1
    backend = instances[0]
    assert backend.seed == 7
    assert backend.generator == "sentinel"
    assert backend.device == "cpu"

    random_result = rng.random(size=5, dtype="float32")
    assert random_result == ("random", 5, "float32", 7)
    assert backend.calls[0] == ("random", 5, "float32")

    uniform_result = rng.uniform(
        low=-1.0,
        high=2.0,
        size=(1, 2),
        dtype="float16",
    )
    assert uniform_result == ("uniform", -1.0, 2.0, (1, 2), "float16", 7)
    assert backend.calls[1] == ("uniform", -1.0, 2.0, (1, 2), "float16")

    normal_result = rng.normal(
        loc=1.5, scale=2.0, size=(2, 2), dtype="float64"
    )
    assert normal_result == ("normal", 1.5, 2.0, (2, 2), "float64", 7)
    assert backend.calls[2] == ("normal", 1.5, 2.0, (2, 2), "float64")

    gamma_result = rng.gamma(
        shape=2.0,
        scale=3.0,
        size=(4,),
        dtype="float32",
    )
    assert gamma_result == ("gamma", 2.0, 3.0, (4,), "float32", 7)
    assert backend.calls[3] == ("gamma", 2.0, 3.0, (4,), "float32")

    choice_result = rng.choice(
        ["a", "b"],
        size=1,
        replace=False,
        p=[0.4, 0.6],
    )
    assert choice_result == ("choice", ["a", "b"], 1, False, [0.4, 0.6], 7)
    assert backend.calls[4] == (
        "choice",
        ["a", "b"],
        1,
        False,
        [0.4, 0.6],
    )


def test_array_rng_rejects_unknown_backend():
    with pytest.raises(ValueError):
        ArrayRNG(backend="unknown")


def test_array_rng_forwards_generator_to_jax(monkeypatch):
    captured_kwargs = {}

    class DummyBackend:
        def __init__(self, **kwargs):
            captured_kwargs.update(kwargs)

        def random(self, *, size, dtype):
            return None

        def normal(self, *, loc, scale, size, dtype):
            return None

        def gamma(self, *, shape, scale, size, dtype):
            return None

        def choice(self, population, *, size, replace, probabilities):
            return None

    monkeypatch.setitem(
        _FACTORIES, "jax", lambda **kwargs: DummyBackend(**kwargs)
    )

    key = ("jax-key",)
    ArrayRNG(backend="jax", generator=key)

    assert captured_kwargs["generator"] is key


def test_numpy_backend_import_error_mentions_extra(monkeypatch):
    # Ensure numPy import attempts go through builtins.__import__
    monkeypatch.delitem(sys.modules, "numpy", raising=False)
    original_import = builtins.__import__

    def fake_import(name, *args, **kwargs):
        if name == "numpy":
            raise ImportError("missing numpy")
        return original_import(name, *args, **kwargs)

    monkeypatch.setattr(builtins, "__import__", fake_import)

    with pytest.raises(ImportError) as err:
        numpy_backend.NumPyBackend(seed=None, generator=None)

    assert "pip install orng[numpy]" in str(err.value)



from __future__ import annotations

from typing import Any

from .._utils import SizeLike, normalize_shape, total_size


class TorchBackend:
    def __init__(
        self,
        *,
        seed: int | None,
        generator: Any | None,
        device: Any | None,
    ) -> None:
        try:
            import torch
        except ImportError as exc:  # pragma: no cover - optional dependency
            raise ImportError(
                "PyTorch backend requires the 'torch' package to be installed. "
                "Install it with `pip install orng[torch]`."
            ) from exc

        self._torch = torch
        self._device = device if device is not None else torch.device("cpu")
        if generator is None:
            self._generator = torch.Generator(device=self._device)
            if seed is not None:
                self._generator.manual_seed(seed)
        elif isinstance(generator, torch.Generator):
            self._generator = generator
        else:
            raise TypeError(
                "generator must be a torch.Generator when using the PyTorch "
                "backend."
            )

    def random(self, *, size: SizeLike, dtype: Any | None) -> Any:
        shape = normalize_shape(size)
        dtype = dtype if dtype is not None else self._torch.float32
        if not shape:
            result = self._torch.rand(
                (1,),
                generator=self._generator,
                device=self._device,
                dtype=dtype,
            )
            return result[0]
        return self._torch.rand(
            shape,
            generator=self._generator,
            device=self._device,
            dtype=dtype,
        )

    def uniform(
        self,
        *,
        low: Any,
        high: Any,
        size: SizeLike,
        dtype: Any | None,
    ) -> Any:
        shape = normalize_shape(size)
        dtype = dtype if dtype is not None else self._torch.float32
        sample_shape = shape if shape else (1,)
        base = self._torch.rand(
            sample_shape,
            generator=self._generator,
            device=self._device,
            dtype=dtype,
        )
        low_tensor = self._torch.as_tensor(
            low, device=self._device, dtype=dtype
        )
        high_tensor = self._torch.as_tensor(
            high, device=self._device, dtype=dtype
        )
        result = low_tensor + (high_tensor - low_tensor) * base
        if not shape:
            return result[0]
        return result

    def normal(
        self,
        *,
        loc: Any,
        scale: Any,
        size: SizeLike,
        dtype: Any | None,
    ) -> Any:
        shape = normalize_shape(size)
        dtype = dtype if dtype is not None else self._torch.float32
        if not shape:
            sample = self._torch.randn(
                (1,),
                generator=self._generator,
                device=self._device,
                dtype=dtype,
            )
            return sample[0] * scale + loc
        base = self._torch.randn(
            shape,
            generator=self._generator,
            device=self._device,
            dtype=dtype,
        )
        return base * scale + loc

    def gamma(
        self,
        *,
        shape: Any,
        scale: Any,
        size: SizeLike,
        dtype: Any | None,
    ) -> Any:
        sample_shape = normalize_shape(size)
        torch = self._torch
        dtype = dtype if dtype is not None else torch.float32
        concentration = torch.as_tensor(
            shape,
            device=self._device,
            dtype=dtype,
        )
        scale_tensor = torch.as_tensor(
            scale,
            device=self._device,
            dtype=dtype,
        )
        concentration, scale_tensor = torch.broadcast_tensors(
            concentration, scale_tensor
        )
        if torch.any(concentration <= 0):
            raise ValueError("shape parameters for gamma must be positive.")
        if sample_shape:
            prefix = (None,) * len(sample_shape)
            concentration = concentration[prefix].expand(
                sample_shape + concentration.shape
            )
            scale_tensor = scale_tensor[prefix].expand(
                sample_shape + scale_tensor.shape
            )
        samples = torch._standard_gamma(
            concentration, generator=self._generator
        )
        scaled = samples * scale_tensor
        return scaled

    def choice(
        self,
        population: int | Any,
        *,
        size: SizeLike,
        replace: bool,
        probabilities: Any | None,
    ) -> Any:
        shape = normalize_shape(size)
        num_samples = total_size(shape)

        if isinstance(population, int):
            pop_size = population
            population_tensor = None
        else:
            population_tensor = self._torch.as_tensor(
                population, device=self._device
            )
            if population_tensor.ndim == 0:
                population_tensor = population_tensor.reshape(1)
            pop_size = int(population_tensor.shape[0])

        if probabilities is not None:
            probs_tensor = self._torch.as_tensor(
                probabilities,
                dtype=self._torch.double,
                device=self._device,
            )
            if probs_tensor.ndim != 1 or probs_tensor.shape[0] != pop_size:
                raise ValueError(
                    "Probabilities must be a 1-D array matching the population."
                )
            indices = self._torch.multinomial(
                probs_tensor,
                num_samples,
                replacement=replace,
                generator=self._generator,
            )
        else:
            if not replace and num_samples > pop_size:
                raise ValueError(
                    "Cannot take a larger sample than population when "
                    "replace=False."
                )
            if replace:
                indices = self._torch.randint(
                    pop_size,
                    (num_samples,),
                    generator=self._generator,
                    device=self._device,
                )
            else:
                indices = self._torch.randperm(
                    pop_size,
                    generator=self._generator,
                    device=self._device,
                )[:num_samples]

        if population_tensor is not None:
            draws = population_tensor.index_select(0, indices)
        else:
            draws = indices

        if not shape:
            return draws[0]
        return draws.reshape(shape)


__all__ = ["TorchBackend"]



from __future__ import annotations

from typing import Any

from .._utils import SizeLike


class NumPyBackend:
    def __init__(self, *, seed: int | None, generator: Any | None) -> None:
        try:
            import numpy as np
        except ImportError as exc:  # pragma: no cover - optional dependency
            raise ImportError(
                "NumPy backend requires the 'numpy' package to be installed. "
                "Install it with `pip install orng[numpy]`."
            ) from exc

        if generator is None:
            self._generator = np.random.default_rng(seed)
        elif isinstance(generator, np.random.Generator):
            self._generator = generator
        else:
            raise TypeError(
                "generator must be a numpy.random.Generator when using the "
                "NumPy backend."
            )

    def random(self, *, size: SizeLike, dtype: Any | None) -> Any:
        return self._generator.random(size=size, dtype=dtype)

    def uniform(
        self,
        *,
        low: Any,
        high: Any,
        size: SizeLike,
        dtype: Any | None,
    ) -> Any:
        result = self._generator.uniform(
            low=low,
            high=high,
            size=size,
        )
        if dtype is not None:
            import numpy as np

            return np.asarray(result, dtype=dtype)
        return result

    def normal(
        self,
        *,
        loc: Any,
        scale: Any,
        size: SizeLike,
        dtype: Any | None,
    ) -> Any:
        result = self._generator.normal(
            loc=loc,
            scale=scale,
            size=size,
        )
        if dtype is not None:
            import numpy as np

            return np.asarray(result, dtype=dtype)
        return result

    def gamma(
        self,
        *,
        shape: Any,
        scale: Any,
        size: SizeLike,
        dtype: Any | None,
    ) -> Any:
        result = self._generator.gamma(
            shape=shape,
            scale=scale,
            size=size,
        )
        if dtype is not None:
            import numpy as np

            return np.asarray(result, dtype=dtype)
        return result

    def choice(
        self,
        population: int | Any,
        *,
        size: SizeLike,
        replace: bool,
        probabilities: Any | None,
    ) -> Any:
        return self._generator.choice(
            population,
            size=size,
            replace=replace,
            p=probabilities,
        )


__all__ = ["NumPyBackend"]



from __future__ import annotations

from typing import Any

from .._utils import SizeLike


class CuPyBackend:
    def __init__(self, *, seed: int | None, generator: Any | None) -> None:
        try:
            import cupy as cp
        except ImportError as exc:  # pragma: no cover - optional dependency
            raise ImportError(
                "CuPy backend requires the 'cupy' package to be installed. "
                "Install it with `pip install orng[cupy]`."
            ) from exc

        self._cupy = cp
        if generator is None:
            self._generator = cp.random.default_rng(seed)
        elif isinstance(generator, cp.random.Generator):
            self._generator = generator
        else:
            raise TypeError(
                "generator must be a cupy.random.Generator when using the "
                "CuPy backend."
            )

    def random(self, *, size: SizeLike, dtype: Any | None) -> Any:
        return self._generator.random(size=size, dtype=dtype)

    def uniform(
        self,
        *,
        low: Any,
        high: Any,
        size: SizeLike,
        dtype: Any | None,
    ) -> Any:
        return self._generator.uniform(
            low=low,
            high=high,
            size=size,
            dtype=dtype,
        )

    def normal(
        self,
        *,
        loc: Any,
        scale: Any,
        size: SizeLike,
        dtype: Any | None,
    ) -> Any:
        standard = self._generator.standard_normal(
            size=size,
            dtype=dtype,
        )
        result = loc + standard * scale
        return result

    def gamma(
        self,
        *,
        shape: Any,
        scale: Any,
        size: SizeLike,
        dtype: Any | None,
    ) -> Any:
        result = self._generator.gamma(
            shape=shape,
            scale=scale,
            size=size,
        )
        if dtype is not None:
            return self._cupy.asarray(result, dtype=dtype)
        return result

    def choice(
        self,
        population: int | Any,
        *,
        size: SizeLike,
        replace: bool,
        probabilities: Any | None,
    ) -> Any:
        # CuPy's Generator lacks choice, so we implement the core semantics.
        cp = self._cupy
        gen = self._generator

        if isinstance(population, int):
            n = population
            values = None
        else:
            values = cp.asarray(population)
            n = values.shape[0]

        target_size = size if size is not None else ()

        probs = None
        if probabilities is not None:
            probs = cp.asarray(probabilities, dtype=cp.float64)
            probs = probs / cp.sum(probs)

        if replace:
            if probs is None:
                indices = gen.integers(0, n, size=target_size)
            else:
                # Inverse-CDF sampling: draw U(0,1) then locate within the CDF.
                cdf = cp.cumsum(probs)
                draws = gen.random(size=target_size)
                indices = cp.searchsorted(cdf, draws)
        else:
            flat_k = (
                1
                if target_size == ()
                else int(cp.prod(cp.asarray(target_size)))
            )
            if probs is None:
                indices = gen.permutation(n)[:flat_k]
            else:
                # Gumbel-top-k trick for weighted sampling without replacement.
                # Ref: https://arxiv.org/abs/1611.00712 (Gumbel-Softmax).
                # Inverse CDF for Gumbel: -log(-log(U)), U~Uniform(0,1)
                gumbels = -cp.log(-cp.log(gen.random(n)))
                keys = cp.log(probs) + gumbels
                # Select the top-k keys
                indices = cp.argpartition(keys, -flat_k)[-flat_k:]
            if target_size != ():
                indices = cp.reshape(indices, target_size)
            else:
                indices = indices[0]

        if values is None:
            return indices
        return values[indices]


__all__ = ["CuPyBackend"]



from __future__ import annotations

from typing import Any, Callable, Dict

from .cupy import CuPyBackend
from .jax import JAXBackend
from .numpy import NumPyBackend
from .torch import TorchBackend

BackendFactory = Callable[..., Any]

_FACTORIES: Dict[str, BackendFactory] = {
    "numpy": lambda *, seed=None, generator=None, device=None: NumPyBackend(
        seed=seed,
        generator=generator,
    ),
    "cupy": lambda *, seed=None, generator=None, device=None: CuPyBackend(
        seed=seed,
        generator=generator,
    ),
    "torch": lambda *, seed=None, generator=None, device=None: TorchBackend(
        seed=seed,
        generator=generator,
        device=device,
    ),
    "pytorch": lambda *, seed=None, generator=None, device=None: TorchBackend(
        seed=seed,
        generator=generator,
        device=device,
    ),
    "jax": lambda *, seed=None, generator=None, device=None: JAXBackend(
        seed=seed,
        key=generator,
    ),
}


def create_backend(
    name: str,
    *,
    seed: int | None,
    generator: Any | None,
    device: Any | None,
):
    try:
        factory = _FACTORIES[name.lower()]
    except KeyError as exc:  # pragma: no cover - defensive
        supported = "', '".join(
            sorted({k for k in _FACTORIES if k != "pytorch"})
        )
        raise ValueError(
            f"Unsupported backend '{name}'. Expected one of '{supported}'."
        ) from exc
    return factory(seed=seed, generator=generator, device=device)


__all__ = ["create_backend"]



"""Backend-aware random number generation helpers.

This module introduces :class:`ArrayRNG`, a small facade that mimics the subset
of ``numpy.random.Generator`` APIs. The class presents a uniform interface
across NumPy, PyTorch, CuPy, and JAX.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Protocol

from ._utils import SizeLike
from .backends import create_backend


class RNGBackend(Protocol):
    """Protocol representing the shim each backend must implement."""

    def random(self, *, size: SizeLike, dtype: Any | None) -> Any: ...

    def uniform(
        self,
        *,
        low: Any,
        high: Any,
        size: SizeLike,
        dtype: Any | None,
    ) -> Any: ...

    def normal(
        self,
        *,
        loc: Any,
        scale: Any,
        size: SizeLike,
        dtype: Any | None,
    ) -> Any: ...

    def gamma(
        self,
        *,
        shape: Any,
        scale: Any,
        size: SizeLike,
        dtype: Any | None,
    ) -> Any: ...

    def choice(
        self,
        population: int | Any,
        *,
        size: SizeLike,
        replace: bool,
        probabilities: Any | None,
    ) -> Any: ...


@dataclass
class ArrayRNG:
    """Facade exposing ``numpy.random.Generator``-style helpers across backends.

    Parameters
    ----------
    backend:
        String identifier for the target library.  Accepted values are
        ``"numpy"``, ``"torch"``/``"pytorch"``, ``"cupy"``, and ``"jax"``.
    seed:
        Optional seed used when constructing a fresh generator.
    generator:
        Optional backend-specific state to wrap.  The expected value depends on
        the backend:

        * NumPy – ``numpy.random.Generator``.
        * PyTorch – ``torch.Generator``.
        * CuPy – ``cupy.random.Generator``.
        * JAX – a ``jax.random.KeyArray`` generated via ``jax.random.key``.

        When supplied, the instance is wrapped instead of creating a new
        generator from ``seed``.
    device:
        Optional device specification forwarded to backends that understand the
        concept (currently PyTorch).
    """

    backend: str = "numpy"
    seed: int | None = None
    generator: Any | None = None
    device: Any | None = None

    def __post_init__(self) -> None:
        self._impl: RNGBackend = create_backend(
            self.backend,
            seed=self.seed,
            generator=self.generator,
            device=self.device,
        )

    # Public API -----------------------------------------------------------------

    def random(
        self,
        size: SizeLike = None,
        *,
        dtype: Any | None = None,
    ) -> Any:
        """Draw samples from ``Uniform[0, 1)``."""
        return self._impl.random(size=size, dtype=dtype)

    def uniform(
        self,
        low: Any = 0.0,
        high: Any = 1.0,
        size: SizeLike = None,
        *,
        dtype: Any | None = None,
    ) -> Any:
        """Draw samples from ``Uniform[low, high)``."""
        return self._impl.uniform(
            low=low,
            high=high,
            size=size,
            dtype=dtype,
        )

    def normal(
        self,
        loc: Any = 0.0,
        scale: Any = 1.0,
        size: SizeLike = None,
        *,
        dtype: Any | None = None,
    ) -> Any:
        """Draw samples from a normal distribution."""
        return self._impl.normal(
            loc=loc,
            scale=scale,
            size=size,
            dtype=dtype,
        )

    def gamma(
        self,
        shape: Any,
        scale: Any = 1.0,
        size: SizeLike = None,
        *,
        dtype: Any | None = None,
    ) -> Any:
        """Draw samples from a gamma distribution."""
        return self._impl.gamma(
            shape=shape,
            scale=scale,
            size=size,
            dtype=dtype,
        )

    def choice(
        self,
        a: int | Any,
        size: SizeLike = None,
        replace: bool = True,
        p: Any | None = None,
    ) -> Any:
        """Sample from a discrete domain."""
        return self._impl.choice(
            a,
            size=size,
            replace=replace,
            probabilities=p,
        )


__all__ = ["ArrayRNG"]



import importlib.metadata

try:
    __version__ = importlib.metadata.version("orng")
except importlib.metadata.PackageNotFoundError:
    __version__ = "unknown"

from .orng import ArrayRNG

__all__ = ["ArrayRNG", "__version__"]



from __future__ import annotations

import math
from typing import Sequence, Tuple

SizeLike = int | Sequence[int] | None


def normalize_shape(size: SizeLike) -> Tuple[int, ...]:
    """Convert a size argument to a canonical ``tuple`` form."""
    if size is None:
        return ()
    if isinstance(size, int):
        if size < 0:
            raise ValueError("size must be non-negative.")
        return (size,)
    shape = tuple(int(dim) for dim in size)
    if any(dim < 0 for dim in shape):
        raise ValueError("size entries must be non-negative.")
    return shape


def total_size(shape: Tuple[int, ...]) -> int:
    if not shape:
        return 1
    return math.prod(shape)


__all__ = ["SizeLike", "normalize_shape", "total_size"]



from __future__ import annotations

import pytest

from orng import ArrayRNG
from orng.backends.cupy import CuPyBackend
from orng.backends.jax import JAXBackend
from orng.backends.numpy import NumPyBackend
from orng.backends.torch import TorchBackend


@pytest.fixture(params=["numpy", "torch", "cupy", "jax"])
def backend_name(request):
    if request.param == "cupy":
        cupy = pytest.importorskip("cupy")
        try:
            cupy.cuda.runtime.getDeviceCount()
        except (ImportError, cupy.cuda.runtime.CUDARuntimeError) as exc:
            pytest.skip(f"CuPy runtime unavailable: {exc}")
    elif request.param == "jax":
        pytest.importorskip("jax")
        pytest.importorskip("jax.numpy")
    elif request.param == "torch":
        pytest.importorskip("torch")
    elif request.param == "numpy":
        pytest.importorskip("numpy")
    else:
        raise ValueError(
            f"Unexpected backend fixture parameter: {request.param}"
        )
    return request.param


@pytest.fixture()
def seed():
    return 42


@pytest.fixture()
def backend_case(backend_name):
    name = backend_name

    if name == "numpy":
        np = pytest.importorskip("numpy")

        def new_backend(seed):
            return NumPyBackend(seed=seed, generator=None)

        def assert_array(arr, shape, dtype):
            assert isinstance(arr, np.ndarray)
            assert tuple(arr.shape) == shape
            if dtype is not None:
                assert arr.dtype == dtype

        def assert_close(a, b):
            np.testing.assert_allclose(a, b)

        def to_list(arr):
            values = arr.tolist()
            return values if isinstance(values, list) else [values]

        return {
            "name": name,
            "seed": 123,
            "new_backend": new_backend,
            "assert_array": assert_array,
            "assert_close": assert_close,
            "to_list": to_list,
            "dtype64": np.float64,
            "dtype32": np.float32,
        }

    if name == "torch":
        torch = pytest.importorskip("torch")

        def new_backend(seed):
            return TorchBackend(seed=seed, generator=None, device=None)

        def assert_array(arr, shape, dtype):
            assert isinstance(arr, torch.Tensor)
            assert tuple(arr.shape) == shape
            if dtype is not None:
                assert arr.dtype == dtype

        def assert_close(a, b):
            assert torch.allclose(a, b)

        def to_list(arr):
            if isinstance(arr, torch.Tensor):
                return arr.cpu().tolist()
            return [arr]

        return {
            "name": name,
            "seed": 123,
            "new_backend": new_backend,
            "assert_array": assert_array,
            "assert_close": assert_close,
            "to_list": to_list,
            "dtype64": torch.float64,
            "dtype32": torch.float32,
        }

    if name == "cupy":
        cp = pytest.importorskip("cupy")
        try:
            cp.cuda.runtime.getDeviceCount()
        except cp.cuda.runtime.CUDARuntimeError as exc:
            pytest.skip(f"CuPy runtime unavailable: {exc}")

        def new_backend(seed):
            return CuPyBackend(seed=seed, generator=None)

        def assert_array(arr, shape, dtype):
            assert isinstance(arr, cp.ndarray)
            assert tuple(arr.shape) == shape
            if dtype is not None:
                assert arr.dtype == dtype

        def assert_close(a, b):
            cp.testing.assert_allclose(a, b)

        def to_list(arr):
            return arr.get().tolist()

        return {
            "name": name,
            "seed": 123,
            "new_backend": new_backend,
            "assert_array": assert_array,
            "assert_close": assert_close,
            "to_list": to_list,
            "dtype64": cp.float64,
            "dtype32": cp.float32,
        }

    if name == "jax":
        jax = pytest.importorskip("jax")
        jnp = pytest.importorskip("jax.numpy")
        array_type = getattr(jax, "Array", type(jnp.asarray(0)))

        def new_backend(seed):
            return JAXBackend(seed=seed, key=None)

        def assert_array(arr, shape, dtype):
            assert isinstance(arr, array_type)
            assert tuple(arr.shape) == shape
            if dtype is not None:
                assert arr.dtype == dtype

        def assert_close(a, b):
            assert bool(jnp.allclose(a, b))

        def to_list(arr):
            values = arr.tolist()
            return values if isinstance(values, list) else [values]

        return {
            "name": name,
            "seed": 123,
            "new_backend": new_backend,
            "assert_array": assert_array,
            "assert_close": assert_close,
            "to_list": to_list,
            "dtype64": jnp.float64,
            "dtype32": jnp.float32,
        }

    raise AssertionError(f"Unexpected backend fixture parameter: {name}")


@pytest.fixture(params=["random", "uniform", "normal", "gamma", "choice"])
def method_case(request):
    if request.param == "random":
        return {
            "method": "random",
            "args": {},
            "kwargs": {"size": (2, 3), "dtype": None},
        }
    elif request.param == "uniform":
        return {
            "method": "uniform",
            "args": {},
            "kwargs": {"low": 0.0, "high": 1.0, "size": (2, 3), "dtype": None},
        }
    elif request.param == "normal":
        return {
            "method": "normal",
            "args": {},
            "kwargs": {
                "loc": 0.0,
                "scale": 1.0,
                "size": (2, 3),
                "dtype": None,
            },
        }
    elif request.param == "gamma":
        return {
            "method": "gamma",
            "args": {},
            "kwargs": {
                "shape": 2.0,
                "scale": 2.0,
                "size": (2, 3),
                "dtype": None,
            },
        }
    elif request.param == "choice":
        return {
            "method": "choice",
            "args": {"a": [0, 1, 2, 3]},
            "kwargs": {"size": (2, 3), "replace": True, "probabilities": None},
        }
    else:
        raise ValueError(
            f"Unexpected method fixture parameter: {request.param}"
        )


def test_backend_seeding(backend_case, method_case):
    seed = backend_case["seed"]
    backend = backend_case["new_backend"](seed)
    clone = backend_case["new_backend"](seed)

    method = getattr(backend, method_case["method"])
    clone_method = getattr(clone, method_case["method"])

    result = method(*method_case["args"].values(), **method_case["kwargs"])
    backend_case["assert_array"](
        result,
        shape=method_case["kwargs"]["size"],
        dtype=method_case["kwargs"].get("dtype", None),
    )

    clone_result = clone_method(
        *method_case["args"].values(), **method_case["kwargs"]
    )
    backend_case["assert_close"](result, clone_result)


@pytest.fixture()
def rng(backend_name, seed):
    return ArrayRNG(backend_name, seed=seed)


def test_array_rng_random(rng):
    x = rng.random(size=(10, 2))
    assert x.shape == (10, 2)


def test_array_rng_uniform(rng):
    x = rng.uniform(low=-1.0, high=1.0, size=(5, 3))
    assert x.shape == (5, 3)


def test_array_rng_normal(rng):
    x = rng.normal(loc=0.0, scale=2.0, size=(4, 4))
    assert x.shape == (4, 4)


def test_array_rng_gamma(rng):
    x = rng.gamma(shape=2.0, scale=2.0, size=(6, 2))
    assert x.shape == (6, 2)

    x = rng.gamma(shape=[1.0, 2.0, 3.0], scale=1.0)
    assert x.shape == (3,)

    x = rng.gamma(shape=2.0, scale=[1.0, 2.0, 3.0])
    assert x.shape == (3,)


def test_array_rng_choice(rng):
    x = rng.choice([10, 20, 30], size=(3, 3), replace=True)
    assert x.shape == (3, 3)
    assert all(v in [10, 20, 30] for v in x.flatten())


def test_cupy_choice_with_probabilities_and_no_replacement():
    cp = pytest.importorskip("cupy")
    try:
        cp.cuda.runtime.getDeviceCount()
    except (ImportError, cp.cuda.runtime.CUDARuntimeError) as exc:
        pytest.skip(f"CuPy runtime unavailable: {exc}")

    backend = CuPyBackend(seed=123, generator=None)

    values = cp.array([1, 2, 3, 4], dtype=cp.int32)
    probs = cp.array([0.7, 0.1, 0.1, 0.1], dtype=cp.float32)

    draws = backend.choice(
        values,
        size=(2, 3),
        replace=True,
        probabilities=probs,
    )
    assert isinstance(draws, cp.ndarray)
    assert draws.shape == (2, 3)
    assert cp.isin(draws, values).all()

    without_replacement = backend.choice(
        values,
        size=(3,),
        replace=False,
        probabilities=None,
    )
    assert isinstance(without_replacement, cp.ndarray)
    assert without_replacement.shape == (3,)
    assert cp.unique(without_replacement).size == 3



from __future__ import annotations

from typing import Any

from .._utils import SizeLike, normalize_shape


class JAXBackend:
    def __init__(self, *, seed: int | None, key: Any | None) -> None:
        try:
            import jax
            from jax import config as jax_config

            if not getattr(jax_config, "x64_enabled", False):
                jax_config.update("jax_enable_x64", True)

            import jax.numpy as jnp
        except ImportError as exc:  # pragma: no cover - optional dependency
            raise ImportError(
                "JAX backend requires the 'jax' package to be installed. "
                "Install it with `pip install orng[jax]`."
            ) from exc

        self._jax = jax
        self._jnp = jnp
        if key is None:
            if seed is None:
                seed = 0
            self._key = jax.random.key(seed)
        else:
            self._key = key

    def _next_key(self) -> Any:
        key, self._key = self._jax.random.split(self._key)
        return key

    def random(self, *, size: SizeLike, dtype: Any | None) -> Any:
        key = self._next_key()
        shape = normalize_shape(size)
        dtype = dtype if dtype is not None else self._jnp.float32
        low = self._jnp.array(0.0, dtype=dtype)
        high = self._jnp.array(1.0, dtype=dtype)
        if shape:
            return self._jax.random.uniform(
                key,
                shape=shape,
                minval=low,
                maxval=high,
                dtype=dtype,
            )
        return self._jax.random.uniform(
            key,
            shape=(1,),
            minval=low,
            maxval=high,
            dtype=dtype,
        )[0]

    def uniform(
        self,
        *,
        low: Any,
        high: Any,
        size: SizeLike,
        dtype: Any | None,
    ) -> Any:
        key = self._next_key()
        shape = normalize_shape(size)
        dtype = dtype if dtype is not None else self._jnp.float32
        low_arr = self._jnp.asarray(low, dtype=dtype)
        high_arr = self._jnp.asarray(high, dtype=dtype)
        if shape:
            return self._jax.random.uniform(
                key,
                shape=shape,
                minval=low_arr,
                maxval=high_arr,
                dtype=dtype,
            )
        return self._jax.random.uniform(
            key,
            shape=(1,),
            minval=low_arr,
            maxval=high_arr,
            dtype=dtype,
        )[0]

    def normal(
        self,
        *,
        loc: Any,
        scale: Any,
        size: SizeLike,
        dtype: Any | None,
    ) -> Any:
        key = self._next_key()
        shape = normalize_shape(size)
        dtype = dtype if dtype is not None else self._jnp.float32
        if shape:
            standard = self._jax.random.normal(
                key,
                shape=shape,
                dtype=dtype,
            )
        else:
            standard = self._jax.random.normal(
                key,
                shape=(1,),
                dtype=dtype,
            )[0]
        return standard * scale + loc

    def gamma(
        self,
        *,
        shape: Any,
        scale: Any,
        size: SizeLike,
        dtype: Any | None,
    ) -> Any:
        key = self._next_key()
        sample_shape = normalize_shape(size)
        dtype = dtype if dtype is not None else self._jnp.float32
        concentration = self._jnp.asarray(shape, dtype=dtype)
        scale_arr = self._jnp.asarray(scale, dtype=dtype)
        # Handle broadcasting of shape and scale
        if sample_shape:
            draw_shape = sample_shape + self._jnp.shape(concentration)
        else:
            draw_shape = self._jnp.shape(concentration)
        gamma_samples = self._jax.random.gamma(
            key,
            concentration,
            shape=draw_shape,
            dtype=dtype,
        )
        scaled = gamma_samples * scale_arr
        if (
            not sample_shape
            and self._jnp.ndim(scale) == 0
            and self._jnp.ndim(shape) == 0
        ):
            return scaled[0]
        return scaled

    def choice(
        self,
        population: int | Any,
        *,
        size: SizeLike,
        replace: bool,
        probabilities: Any | None,
    ) -> Any:
        key = self._next_key()
        shape = normalize_shape(size)
        jax_shape = shape if shape else None
        if isinstance(population, int):
            domain = population
        else:
            domain = self._jnp.asarray(population)
        probs = (
            None if probabilities is None else self._jnp.asarray(probabilities)
        )
        result = self._jax.random.choice(
            key,
            domain,
            shape=jax_shape,
            replace=replace,
            p=probs,
        )
        if not shape:
            return result
        return result


__all__ = ["JAXBackend"]
